/*
 * ParserEventHandlerImpl.java
 * Created on 2011/10/13
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

package jp.co.ntt.oss.heapstats.gui.snapshot;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import jp.co.ntt.oss.heapstats.container.ObjectData;
import jp.co.ntt.oss.heapstats.container.SnapShot;
import jp.co.ntt.oss.heapstats.container.SnapShotHeader;
import jp.co.ntt.oss.heapstats.gui.GuiCommon;
import jp.co.ntt.oss.heapstats.parser.ParserEventHandler;

/**
 * This class is written to the temporary file information file java heap.
 */
public class ParserEventHandlerImpl implements
                                   ParserEventHandler {
    /**
     * Holds the absolute path to the directory that contains a snapshot of one.
     */
    private String tmpDirName;

    /** Hold the data in the snapshot. */
    private SnapShot snapShot;

    /**
     * Key is snapshot date. Value is the absolute path to the directory that
     * contains a snapshot of one.
     */
    private Map<Date, String> snapShotList;

    /**
     * This constructor initializes the variable.
     */
    public ParserEventHandlerImpl() {
        snapShotList = new TreeMap<Date, String>();
    }

    @Override
    public final ParseResult onStart(final long off) {
        snapShot = new SnapShot();
        snapShot.setElements(new HashMap<Long, ObjectData>());
        snapShot.setToChildren(new HashMap<Long, List<ObjectData>>());
        snapShot.setToParents(new HashMap<Long, List<ObjectData>>());

        return ParseResult.HEAPSTATS_PARSE_CONTINUE;
    }

    @Override
    public final ParseResult onNewSnapShot(final SnapShotHeader header,
            final String parent) {
        ParseResult value = ParseResult.HEAPSTATS_PARSE_CONTINUE;
        SimpleDateFormat dateFmt = new SimpleDateFormat("yyyyMMddHHmmssSSS");
        File parentFile = new File(parent);
        StringBuilder buf = new StringBuilder();

        buf.append("snapshot");
        buf.append(File.separator);
        if (parentFile.getParent() != null) {
            buf.append(parentFile.getParentFile().getName());
        } else {
            buf.append("tmp");
        }
        buf.append(File.separator);
        buf.append(parentFile.getName());
        buf.append(File.separator);
        buf.append(dateFmt.format(header.getSnapShotDate()));

        tmpDirName = buf.toString();

        deleteReferenceFile();

        if (writeHeader(header)) {
            snapShotList.put(header.getSnapShotDate(), buf.toString());
        } else {
            value = ParseResult.HEAPSTATS_PARSE_ABORT;
        }

        return value;
    }

    @Override
    public final ParseResult onEntry(final ObjectData data) {
        snapShot.getElements().put(data.getTag(), data);
        return ParseResult.HEAPSTATS_PARSE_CONTINUE;
    }

    @Override
    public final ParseResult onChildEntry(final long parentClassTag,
            final long classTag, final long instances, final long totalSize) {
        List<ObjectData> toChild = snapShot.getToChildren().get(parentClassTag);
        if (toChild == null) {
            toChild = new LinkedList<ObjectData>();
            snapShot.getToChildren().put(parentClassTag, toChild);
        }
        ObjectData child = new ObjectData();
        child.setTag(classTag);
        child.setCount(instances);
        child.setTotalSize(totalSize);

        toChild.add(child);

        List<ObjectData> toParent = snapShot.getToParents().get(classTag);
        if (toParent == null) {
            toParent = new LinkedList<ObjectData>();
            snapShot.getToParents().put(classTag, toParent);
        }

        ObjectData parent = new ObjectData();
        parent.setTag(parentClassTag);
        parent.setCount(instances);
        parent.setTotalSize(totalSize);

        toParent.add(parent);

        return ParseResult.HEAPSTATS_PARSE_CONTINUE;
    }

    @Override
    public final ParseResult onFinish(final long off) {
        if (!writeClassEntry()) {
            return ParseResult.HEAPSTATS_PARSE_ABORT;
        }
        if (!writeReference()) {
            return ParseResult.HEAPSTATS_PARSE_ABORT;
        }

        return ParseResult.HEAPSTATS_PARSE_CONTINUE;
    }

    /**
     * Return to map the data in the snapshot.
     *
     * @return Return to map the data in the snapshot
     */
    public final Map<Date, String> getSnapShotList() {
        return snapShotList;
    }

    /**
     * Write to temp files header information of the snapshot.
     *
     * @param header Header information
     * @return Returns the result of a write.
     */
    private boolean writeHeader(final SnapShotHeader header) {
        File tmpFile = new File(tmpDirName);
        if (!tmpFile.exists()) {
            if (!tmpFile.mkdirs()) {
                return false;
            }
        }
        tmpFile.deleteOnExit();

        File tmpHeader = new File(tmpDirName + File.separator
                + GuiCommon.SNAPSHOT_TEMP_NAME_HEADER);
        tmpHeader.deleteOnExit();
        if (tmpHeader.exists()) {
            return true;
        }

        PrintWriter write = null;
        try {
            write = new PrintWriter(tmpHeader);

            write.write(String.valueOf(header.getSnapShotDate().getTime()));
            write.write(",");
            write.write(String.valueOf(header.getNumEntries()));
            write.write(",");
            write.write(String.valueOf(header.getCause()));
            write.write(",");
            write.write(String.valueOf(header.getGcCause()));
            write.write(",");
            write.write(String.valueOf(header.getFullCount()));
            write.write(",");
            write.write(String.valueOf(header.getYngCount()));
            write.write(",");
            write.write(String.valueOf(header.getGcTime()));
            write.write(",");
            write.write(String.valueOf(header.getNewHeap()));
            write.write(",");
            write.write(String.valueOf(header.getOldHeap()));
            write.write(",");
            write.write(String.valueOf(header.getTotalSize()));
            write.write(",");
            write.write(String.valueOf(header.getMetaspaceUsage()));
            write.write(",");
            write.write(String.valueOf(header.getMetaspaceCapacity()));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            return false;
        } finally {
            if (write != null) {
                write.close();
            }
        }

        return true;
    }

    /**
     * Written to a temporary file class information.
     *
     * @return Returns the result of a write.
     */
    private boolean writeClassEntry() {
        File tmpFile = new File(tmpDirName + File.separator
                + GuiCommon.SNAPSHOT_TEMP_NAME_CLASS);
        tmpFile.deleteOnExit();
        if (tmpFile.exists()) {
            return true;
        }

        PrintWriter write = null;
        try {
            write = new PrintWriter(tmpFile);

            for (Map.Entry<Long, ObjectData> entry : snapShot.getElements()
                    .entrySet()) {
                write.write(entry.getValue().toString());
                write.println();
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            return false;
        } finally {
            if (write != null) {
                write.close();
            }
        }
        return true;
    }

    /**
     * Written to a temporary file to the child or parent class information.
     *
     * @return Returns the result of a write
     */
    private boolean writeReference() {
        FileWriter write = null;
        File tmpFile = null;
        try {
            StringBuilder path = new StringBuilder();
            path.append(tmpDirName);
            path.append(File.separator);
            path.append(GuiCommon.SNAPSHOT_TEMP_NAME_REFERENCE);

            tmpFile = new File(path.toString());
            tmpFile.deleteOnExit();

            write = new FileWriter(tmpFile);

            // child
            write(write, snapShot.getToChildren(), true);
            // parent
            write(write, snapShot.getToParents(), false);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        } finally {
            if (write != null) {
                try {
                    write.close();
                } catch (IOException e) {
                    e.printStackTrace();
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Writes the information to a file of the class to which it refers.
     *
     * @param write Convenience class for writing character files
     * @param objectDataMap Child or parent class information
     * @param isChild If the child class to set the true
     * @throws Exception If an error has occurred.
     */
    private void write(final FileWriter write,
            final Map<Long, List<ObjectData>> objectDataMap,
            final boolean isChild) throws Exception {
        for (Map.Entry<Long, List<ObjectData>> tmp : objectDataMap.entrySet()) {
            ObjectData keyData = snapShot.getElements().get(tmp.getKey());
            if (keyData == null) {
                continue;
            }
            for (ObjectData entry : tmp.getValue()) {
                write.write(String.valueOf(keyData.getTag()));
                write.write(",");
                if (isChild) {
                    write.write(GuiCommon.SNAPSHOT_TEMP_CHILD);
                } else {
                    write.write(GuiCommon.SNAPSHOT_TEMP_PARENT);
                }

                write.write(",");
                write.write(String.valueOf(entry.getTag()));
                write.write(",");
                write.write(String.valueOf(entry.getCount()));
                write.write(",");
                write.write(String.valueOf(entry.getTotalSize()));
                write.write(",");

                ObjectData loaderObj = snapShot.getElements().get(
                        entry.getTag());
                long loader = -1;
                if (loaderObj != null) {
                    loader = loaderObj.getClassLoaderTag();
                }

                if (loader < 0) {
                    write.write("-");
                } else if (loader == 0) {
                    write.write("<SystemClassLoader>");
                } else {
                    ObjectData loaderInfo = snapShot.getElements().get(loader);
                    if (loaderInfo != null) {
                        write.write(loaderInfo.getName());
                    } else {
                        write.write("-");
                    }

                    // ClassLoader ID
                    write.write("(0x"
                            + Long.toHexString(loaderObj.getClassLoader())
                            + ")");
                }
                write.write("\n");
            }
        }
    }

    /**
     * Delete the temporary files in the snapshot.
     *
     * @return Returns the result of deleting the file.
     */
    private boolean deleteReferenceFile() {
        boolean ret = true;

        File file = new File(tmpDirName);
        File[] listFiles = file.listFiles();
        if (listFiles != null) {
            for (File entry : listFiles) {
                if (!entry.delete()) {
                    break;
                }
            }
        }
        return ret;
    }
}
